import json
import re
import os
import pandas as pd
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

matplotlib.use('TkAgg')


def plot_benchmark(monitoring_log, parent_workflow_name, cromwell_id, output_dir):
    """
    plots the converted/formatted monitoring log (JSON)
    :param monitoring_log: monitoring log in JSON format
    :param parent_workflow_name: the name of the workflow being run
    :param cromwell_id: the cromwell id for the workflow being run
    :param output_dir: the desired output directory to save the plot
    """

    with open(monitoring_log, "r") as ml:
        calls_dict = json.load(ml)

    plot_memory_usage(calls_dict, parent_workflow_name, cromwell_id, output_dir)


def plot_memory_usage(calls_dict, parent_workflow_name, cromwell_id, output_dir):
    """
    plots a graph of the max cpu, disk and mem usage for each call
    :param calls_dict: dictionary data structure of monitoring log
    :param parent_workflow_name: the name of the workflow being run
    :param cromwell_id: the cromwell id for the workflow being run
    :param output_dir: the desired output directory to save the plot
    """

    tasks = {}
    for key, calls in calls_dict.items():
        for call in calls:
            # get the id/name of each task call
            name = str(call["id"])
            # if there the task call is an attempt
            match = re.findall(r"(attempt-\d+)$", name)
            if len(match) > 0:
                # remove attempt from the task call name
                name = os.path.dirname(name)

            match = re.findall(r"(shard-\d+)$", name)
            # if task is not a shard
            if len(match) == 0:
                # get the name of the task
                # i.e: os.path.basename(/workflow/cromwell_id/call) == call
                task_name = os.path.basename(name)

            # if task is a shard
            else:
                # get the name of the task that was scattered plus the shard
                # i.e: os.path.dirname(/workflow/cromwell_id/call/shard) == /workflow_/cromwell_id/call/
                task_name = os.path.basename(os.path.dirname(name))

            # if no other shards of the task have been stored in dict
            if task_name not in tasks:
                # store total memory and max memory of current task call
                total_memory = str(call["summary"]["memory"]["size"]) + str(
                    call["summary"]["memory"]["unit"]
                )
                tasks[task_name] = {"max_memory": 0.0, "total_memory": total_memory}

            # get max memory usage of current task call
            max_memory = 0.0
            logs = call["logs"]
            for log in logs:
                memory_usage = log["memory_usage"]
                if memory_usage > max_memory:
                    max_memory = memory_usage

            # if max memory usage of current task call is > than max usage from other shards
            if max_memory > tasks[task_name]["max_memory"]:
                # save total memory and max memory of current task call
                tasks[task_name]["max_memory"] = max_memory
                tasks[task_name]["total_memory"] = str(
                    call["summary"]["memory"]["size"]
                ) + str(call["summary"]["memory"]["unit"])

    task_calls = []
    total_memories = []
    max_memory_usages = []
    # store in arrays to put in data frame for plotting
    for name, mem_info in tasks.items():
        task_calls.append(name)
        max_memory_usages.append(mem_info["max_memory"])
        total_memories.append(mem_info["total_memory"])

    # create data frame / format for plotting
    data_frame = pd.DataFrame(
        {
            "task calls": task_calls,
            "total memories": total_memories,
            "max memory usages": max_memory_usages,
        }
    )

    if len(task_calls) > 0:
        data_frame = data_frame.sort_values("task calls")

    # plot figure for mem usage
    fig = plt.figure()
    fig.tight_layout()
    fig.suptitle("Memory Usage of " + parent_workflow_name + " '" + cromwell_id + "'")
    bars_index = np.arange(len(task_calls))

    # set up left axis
    ax1 = fig.add_subplot(111)
    ax1.barh(bars_index, data_frame["max memory usages"])
    ax1.set_yticks(bars_index)
    ax1.set_yticklabels(data_frame["task calls"])
    ax1.set_xlim([0, 100])
    ax1.tick_params(axis="both", labelsize=4)
    ax1.set_xlabel("Max Memory Usage (%)")

    # set up right axis
    ax2 = ax1.twinx()
    ax2.barh(bars_index, data_frame["max memory usages"])
    ax2.set_yticks(bars_index)
    ax2.set_yticklabels(data_frame["total memories"])
    ax2.set_ylabel("Total Memory Available")
    ax2.yaxis.set_label_position("right")
    ax2.tick_params(axis="both", labelsize=4)

    # save plot
    filename = get_path(
        parent_workflow_name + "_" + cromwell_id + "_memory_usage_plot.png", output_dir
    )
    plt.savefig(filename, dpi=1000, bbox_inches="tight")


def get_path(filename, output_dir):
    """
    :param filename: the filename to be saved in the output directory
    :param output_dir: the desired output directory to store the monitoring log, plots and json file
    :return: returns the combined path of file name and  output directory
    """

    # get the combined path of output dir path, filename and sub dir path
    path = os.path.join(output_dir, filename)

    # if the combined path exist
    if os.path.isfile(path):
        # delete that current path
        os.remove(path)

    # return combined path
    return path



import os
import tempfile
import json
import six
import pytest
import requests
from cromwell_tools.cromwell_auth import CromwellAuth


six.add_move(six.MovedModule('mock', 'mock', 'unittest.mock'))
from six.moves import mock  # noqa


def setup_auth_types():
    temp_dir = tempfile.mkdtemp()
    secrets_file = os.path.join(temp_dir, 'fake_secrets.json')
    service_account_key = os.path.join(temp_dir, 'fake_key.json')
    username = "fake_user"
    password = "fake_password"
    url = "https://fake_url"

    auth_params = {"url": url, "username": username, "password": password}
    with open(secrets_file, 'w') as f:
        json.dump(auth_params, f)

    auth_params['secrets_file'] = {"secrets_file": secrets_file}
    auth_params['service_account_key'] = {
        "service_account_key": service_account_key,
        "url": url,
    }
    return auth_params


auth_types = setup_auth_types()


@mock.patch('cromwell_tools.cromwell_auth.CromwellAuth.from_service_account_key_file')
def test_harmonize_credentials_only_takes_one_auth_type(mock_header):
    url = 'https://cromwell.server.org'
    expected_auth = CromwellAuth(
        url=url, header={"Authorization": "bearer fake_token"}, auth=None
    )
    mock_header.return_value = expected_auth
    with pytest.raises(ValueError):
        CromwellAuth.harmonize_credentials(**auth_types)


def test_harmonize_credentials_user_password():
    username = 'fake_user'
    password = 'fake_password'
    url = 'https://cromwell.server.org'
    expected_auth = CromwellAuth(
        url=url, header=None, auth=requests.auth.HTTPBasicAuth(username, password)
    )
    auth = CromwellAuth.harmonize_credentials(
        username=username, password=password, url=url
    )
    assert auth.auth == expected_auth.auth
    assert auth.header == expected_auth.header


def test_harmonize_credentials_from_secrets_file():
    username = "fake_user"
    password = "fake_password"
    url = "https://fake_url"
    expected_auth = CromwellAuth(
        url=url, header=None, auth=requests.auth.HTTPBasicAuth(username, password)
    )
    auth = CromwellAuth.harmonize_credentials(
        secrets_file=auth_types['secrets_file']['secrets_file']
    )
    assert auth.auth == expected_auth.auth
    assert auth.header == expected_auth.header


@mock.patch('cromwell_tools.cromwell_auth.CromwellAuth.from_service_account_key_file')
def test_harmonize_credentials_from_service_account_key(mock_header):
    service_account_key = 'fake_key.json'
    url = 'https://cromwell.server.org'
    expected_auth = CromwellAuth(
        url=url, header={"Authorization": "bearer fake_token"}, auth=None
    )
    mock_header.return_value = expected_auth
    auth = CromwellAuth.harmonize_credentials(
        url=url, service_account_key=service_account_key
    )
    assert auth == expected_auth


@mock.patch('cromwell_tools.cromwell_auth.CromwellAuth.from_service_account_key_file')
def test_harmonize_credentials_from_service_account_key_content(mock_header):
    service_account_key = {'client_email': 'fake_email', 'token_uri': 'fake_uri'}
    url = 'https://cromwell.server.org'
    expected_auth = CromwellAuth(
        url=url, header={"Authorization": "bearer fake_token"}, auth=None
    )
    mock_header.return_value = expected_auth
    auth = CromwellAuth.harmonize_credentials(
        url=url, service_account_key=service_account_key
    )
    assert auth == expected_auth


def test_harmonize_credentials_from_no_authentication():
    url = "https://fake_url"
    expected_auth = CromwellAuth(url=url, header=None, auth=None)
    auth = CromwellAuth.harmonize_credentials(url=url)
    assert auth.auth == expected_auth.auth
    assert auth.header == expected_auth.header



"""
TODO: add some module docs
TODO: once switched to support only Py3.7+, replace all 'cls'
type annotations with the actual Types, rather than using the strings.
This in Py3.6(-) is limited by the lack of Postponed Evaluation of Annotations, see:
https://www.python.org/dev/peps/pep-0563/
"""

import json
import requests
import requests.auth
import six
from google.oauth2 import service_account
import google.auth.transport.requests
import logging
from typing import Union, Dict

logger = logging.getLogger(__name__)


class CromwellAuth:
    def __init__(
        self,
        url: str,
        header: Dict[str, str] = None,
        auth: requests.auth.HTTPBasicAuth = None,
        oauth_credentials: service_account.Credentials = None,
        service_key_content: dict = None,
    ) -> None:
        """Authentication Helper Class for a Cromwell Server.

        Currently this class only supports 3 AuthN methods with Cromwell:
            - OAuth2 through Google by passing a service account JSON key file.
            - HTTPBasic Authentication.
            - No Auth

        Args:
            url: The URL to the Cromwell server. e.g. "https://cromwell.server.org/"
            header: Dictionary containing the (bearer token) authorization header.
            auth: HTTP Basic Authentication information,
                i.e. username and password.
            oauth_credentials: Optional, service_account.Credentials object,
            used for refreshing the Authentication headers when using OAuth.
            service_key_content: Optional, required by JES-backend Cromwells that use OAuth.
        """

        # TODO: add a step to validate the auth information with Cromwell Server, requires /auth endpoint from Cromwell
        self._validate_auth(header, auth)
        self._header = header
        self._auth = auth
        self._credentials = oauth_credentials
        self.url = self._validate_url(url)
        self._service_key_content = service_key_content

    @property
    def auth(self):
        """The property of HTTP Basic Authentication information, i.e. username and password."""
        return self._auth

    @property
    def service_key_content(self):
        """The content of the service account key, if applicable. This is required by JES-backend Cromwell using OAuth."""
        return self._service_key_content

    @property
    def header(self):
        """Dictionary containing the (bearer token) authorization header.

        If the credentials are provided to create the object, this property will make sure the token is not expired.
        """
        if self._credentials:
            if not self._credentials.valid:
                self._credentials.refresh(google.auth.transport.requests.Request())
                header = {}
                self._credentials.apply(header)
                self._header = header
        return self._header

    @staticmethod
    def _validate_auth(
        header: Union[Dict[str, str], None],
        auth: Union[requests.auth.HTTPBasicAuth, None],
    ) -> None:
        """Validate the format of the auth header and the HTTPBasic Auth credentials.

        Args:
            header: Dictionary containing the (bearer token) authorization header.
            auth: HTTP Basic Authentication information, i.e. username and password.

        Raises:
            TypeError: when header is not a valid `dict` or the auth object is not a valid
                `requests.auth.HTTPBasicAuth` object.
            ValueError: when the header is not a valid header(with Bearer token).
        """
        if not header and not auth:
            logger.warning(
                'You are not using any authentication with Cromwell. For security purposes, '
                'please consider adding authentication in front of your Cromwell instance!'
            )

        if header:
            if not isinstance(header, dict):
                raise TypeError('If passed, header must be a dict.')
            if "authorization" not in [k.lower() for k in header.keys()]:
                raise ValueError(
                    'The header must have an "Authorization" or "authorization" key!'
                )

        if auth and not isinstance(auth, requests.auth.HTTPBasicAuth):
            raise TypeError(
                'The auth object must be a valid "requests.auth.HTTPBasicAuth" object!'
            )

    @staticmethod
    def _validate_url(url: str) -> str:
        """Validate the input Cromwell url and remove the trailing slash.

        Args:
            url: The URL to the Cromwell server. e.g. "https://cromwell.server.org/"

        Returns:
            The URL to the Cromwell server without slash at the end. e.g. "https://cromwell.server.org"

        Raises:
            ValueError: If the input url is invalid, i.e. not following http(s) schema.
        """
        if isinstance(url, six.string_types) and url.startswith('http'):
            return str(url.strip('/'))
        else:
            raise ValueError("url must be an str pointing to an http(s) endpoint.")

    @classmethod
    def from_service_account_key_file(
        cls: 'CromwellAuth', service_account_key: Union[str, dict], url: str
    ) -> 'CromwellAuth':
        """Generate an authentication object from a Service Account JSON key file.

        The service account key file will be required if you are using Cromwell-as-a-Service or
        you are using OAuth for your Cromwell Server.

        Args:
            service_account_key: Path to the JSON key file(service account key), or a dict of
                the JSON content of the key file, for authenticating with CaaS.
            url: The URL to the Cromwell server. e.g. "https://cromwell.server.org/"

        Returns:
            An instance of this auth helper class with valid OAuth Auth header.
        """
        scopes = ['email', 'openid', 'profile']
        if isinstance(service_account_key, dict):
            credentials = service_account.Credentials.from_service_account_info(
                service_account_key, scopes=scopes
            )
            service_key_content = service_account_key
        else:
            credentials = service_account.Credentials.from_service_account_file(
                service_account_key, scopes=scopes
            )
            with open(service_account_key, 'r') as f:
                service_key_content = json.load(f)

        if not credentials.valid:
            credentials.refresh(google.auth.transport.requests.Request())
        header = {}
        credentials.apply(header)
        return cls(
            url=url,
            header=header,
            auth=None,
            oauth_credentials=credentials,
            service_key_content=service_key_content,
        )

    @classmethod
    def from_secrets_file(cls: 'CromwellAuth', secrets_file: str) -> 'CromwellAuth':
        """Generate an authentication object from a JSON file that contains credentials info for HTTPBasicAuth.

        Args:
            secrets_file: Path to the JSON file containing username, password, and url fields. e.g.
                {
                    "username": "",
                    "password": "",
                    "url": ""
                }

        Returns:
            An instance of this auth helper class with valid u/p and url for HTTPBasicAuth.
        """
        with open(secrets_file, 'r') as f:
            secrets = json.load(f)
        auth = requests.auth.HTTPBasicAuth(secrets['username'], secrets['password'])
        url = secrets['url']
        return cls(url=url, header=None, auth=auth)

    @classmethod
    def from_user_password(
        cls: 'CromwellAuth', username: str, password: str, url: str
    ) -> 'CromwellAuth':
        """Generate an authentication object from username, password, and Cromwell url.

        Args:
            username: Cromwell username for HTTPBasicAuth.
            password: Cromwell password for HTTPBasicAuth.
            url: The URL to the Cromwell server. e.g. "https://cromwell.server.org/"

        Returns:
            An instance of this auth helper class with valid u/p and url for HTTPBasicAuth.
        """
        auth = requests.auth.HTTPBasicAuth(username, password)
        return cls(url=url, header=None, auth=auth)

    @classmethod
    def from_no_authentication(cls: 'CromwellAuth', url: str) -> 'CromwellAuth':
        """Generate an authentication object which does use any auth methods.

        Args:
            url: The URL to the Cromwell server. e.g. "https://cromwell.server.org/"

        Returns:
            An instance of this auth helper class with url and has no auth methods.
        """
        return cls(url=url, header=None, auth=None)

    @classmethod
    def harmonize_credentials(
        cls: 'CromwellAuth',
        username: str = None,
        password: str = None,
        url: str = None,
        secrets_file: str = None,
        service_account_key: str = None,
    ) -> 'CromwellAuth':
        """Parse and harmonize user inputted credentials and generate proper authentication object for cromwell-tools.

        Args:
            username: Cromwell username for HTTPBasicAuth.
            password: Cromwell password for HTTPBasicAuth.
            url: The URL to the Cromwell server. e.g. "https://cromwell.server.org/"
            secrets_file: Path to the JSON file containing username, password, and url fields. e.g.
                {
                    "username": "",
                    "password": "",
                    "url": ""
                }
            service_account_key: Path to the JSON key file(service account key) for authenticating
                with CaaS or any Cromwell instances that using OAuth.

        Returns:
            An instance of this auth helper class with proper credentials for authenticating with
                Cromwell or CaaS.
        """
        # verify only one credential provided
        credentials = {
            "service_account_key": all((service_account_key, url)),
            "secrets_file": True if secrets_file else False,
            "user_password": all((username, password, url)),
            "no_auth": True
            if (
                not any((service_account_key, secrets_file, username, password)) and url
            )
            else False,
        }
        if sum(credentials.values()) != 1:
            raise ValueError(
                "Exactly one set of credentials must be passed.\nCredentials: {}".format(
                    repr(credentials)
                )
            )

        if credentials["service_account_key"]:
            return cls.from_service_account_key_file(service_account_key, url)
        if credentials["secrets_file"]:
            return cls.from_secrets_file(secrets_file)
        if credentials["user_password"]:
            return cls.from_user_password(username, password, url)
        if credentials["no_auth"]:
            return cls.from_no_authentication(url=url)



import io
import json
import os
import re
import requests
import shutil
import warnings
import zipfile
from typing import List, Union, Dict, Match, Any
from cromwell_tools.cromwell_auth import CromwellAuth


# Note: the following rules for validating labels were originally based on Cromwell's documentation on Github:
# https://github.com/broadinstitute/cromwell/blob/32/CHANGELOG.md
# However, from Cromwell v32, most of the restrictions on the labels have been moved, according to
# https://cromwell.readthedocs.io/en/stable/Labels/, below are the requirements for a valid label
#  key/value pair in Cromwell:
#
# - Label keys may not be empty but label values may be empty.
# - Label key and values have a max char limit of 255.

_CROMWELL_LABEL_LENGTH = 63
_CROMWELL_LABEL_KEY_REGEX = '[a-z]([-a-z0-9]*[a-z0-9])?'
_CROMWELL_LABEL_VALUE_REGEX = '([a-z0-9]*[-a-z0-9]*[a-z0-9])?'


def _emulate_python_fullmatch(
    regex: str, string: str, flags: Union[int, str] = 0
) -> Union[Match[str], None]:
    """Backport Python 3.4's regular expression "fullmatch()" to Python 2 by emulating python-3.4 re.fullmatch().

    If the whole string matches the regular expression pattern, return a corresponding match object. Return None
    if the string does not match the pattern; note that this is different from a zero-length match.

    Args:
        regex: A regex string.
        string: The string that you want to apply regex match to.
        flags: The expression's behaviour can be modified by specifying a flags value. Values can be any of
            the variables listed in https://docs.python.org/3/library/re.html

    Returns:
        A matched object, or None if the string does not match the pattern.
    """
    return re.match("(?:" + regex + r")\Z", string, flags=flags)


if "fullmatch" not in dir(re):  # For Python3.4+
    re.fullmatch = _emulate_python_fullmatch


def download_to_map(urls: List[str]) -> Dict[str, Any]:
    """Reads contents from each url into memory and returns a map of urls to their contents.

    Args:
        urls: A list of urls to the contents to be downloaded.

    Returns:
        url_to_contents: A dict representing the mapping from url to the downloaded contents in-memory.
    """
    url_to_contents = {}
    for url in urls:
        contents = download(url)
        url_to_contents[url] = contents
    return url_to_contents


def make_zip_in_memory(url_to_contents: Dict[str, Any]) -> io.BytesIO:
    """Given a map of urls and their contents, returns an in-memory zip file containing each file.

    For each url, the part after the last slash is used as the file name when writing to the zip archive.

    Args:
        url_to_contents: A dict representing the mapping from url to the downloaded contents in-memory.

    Returns:
        bytes_buf: Zipped files content in bytes.
    """
    buf = io.BytesIO()
    with zipfile.ZipFile(buf, 'w') as zip_buffer:
        for url, contents in url_to_contents.items():
            name = url.split('/')[-1]
            zip_buffer.writestr(name, contents)

    # To properly send the zip to Cromwell in start_workflow, we need it to be an io.BytesIO.
    # If we don't convert, start_workflow appears to succeed, but Cromwell can't find
    # anything in the zip.
    # (six.BytesIO is just an alias for StringIO in Python 2 and for BytesIO in Python 3.)
    bytes_buf = io.BytesIO(buf.getvalue())
    return bytes_buf


def download(url: str) -> Union[str, bytes]:
    """Reads the contents located at the url into memory and returns them.

    Urls starting with http are fetched with an http request. All others are assumed to be local file paths
    and read from the local file system.

    Args:
        url: The url to the content to be downloaded, or the path to the local file.

    Returns:
        Downloaded content in str or bytes format.

    Raises:
        TypeError: If the url is not a str type.
    """
    if not isinstance(url, str):
        raise TypeError('The url/path must be a (str) type, not {}!'.format(type(url)))

    if url.startswith('http'):
        return download_http(url)
    else:
        return read_local_file(url)


def download_http(url: str) -> Union[str, bytes]:
    """
    Makes an http request for the contents at the given url and returns the response body.

    Args:
        url: The url to the content to be downloaded.

    Returns:
        response_str: Content returned from the server. Will be `str` in Python2 and bytes in `Python3`.
    """
    response = requests.get(url)
    response.raise_for_status()

    # Encoding here prevents a UnicodeDecodeError later in make_zip_in_memory in Python 2.
    response_str = response.text.encode('utf-8')
    return response_str


def read_local_file(path: str) -> Union[str, bytes]:
    """Reads the file contents and returns them.

    Args:
        path: Path to the local file to be loaded.

    Returns:
        contents: The loaded content. bytes in Python3 and str in Python2.
    """
    with open(os.path.abspath(path), 'rb') as f:
        contents = f.read()
    return contents


def _localize_file(url: str, target_directory: str = '.') -> None:
    """Localize file url to a directory. Supports both local files and http(s) endpoints.

    Args:
        url: URL of local or http target.
        target_directory: Directory to localize file to.
    """
    if not os.path.isdir(target_directory):
        raise NotADirectoryError(
            'target_directory must be a valid directory on the local filesystem'
        )

    basename = os.path.basename(url)
    target_file = os.path.join(target_directory, basename)

    if url.startswith('http'):
        data = download_http(url)
        with open(target_file, 'wb') as f:
            f.write(data)
    else:
        if not os.path.isfile(url):
            raise FileNotFoundError(
                'non-http files must point to a valid file on the local filesystem. Not found: {}'.format(
                    url
                )
            )
        else:
            shutil.copy(url, target_file)


def _content_checker(regex: str, content: str) -> str:
    """Helper function to check if a string is obeying the rule described by a regex string or not.

    Args:
        regex: A regex string defines valid content.
        content: A string to be validated.

    Returns:
        A string of error message if validation fails, or an empty string if validation succeeds.
    """
    matched = re.fullmatch(regex, content)

    if not matched:
        return f'Invalid label: {content} does not match the regex {regex}.\n'
    else:
        return ''


def _length_checker(length: int, content: str) -> str:
    """Helper function to check if a string is shorter than expected length of not.

    Args:
        length: Maximum length of an expected string.
        content: A string to be validated.

    Returns:
        A string of error message if validation fails, or an empty string if validation succeeds.
    """
    if len(content) > length:
        return f'Invalid label: {content} has {len(content)} characters. The maximum is {length}.\n'
    else:
        return ''


def validate_cromwell_label(
    label_object: Union[str, io.BytesIO, bytes, Dict[str, str]]
) -> None:
    """Check if the label object is valid for Cromwell.

    Note: this function as well as the global variables _CROMWELL_LABEL_LENGTH, _CROMWELL_LABEL_KEY_REGEX
    and _CROMWELL_LABEL_VALUE_REGEX are implemented based on the Cromwell's documentation:
    https://cromwell.readthedocs.io/en/develop/Labels/ and the Cromwell's code base:
    https://github.com/broadinstitute/cromwell/blob/master/core/src/main/scala/cromwell/core/labels/Label.scala#L16
    Both the docs and the code base of Cromwell could possibly change in the future, please update this
    checker on demand.

    Args:
        label_object: A dictionary or a key-value object string defines a Cromwell label.

    Raises:
        ValueError: This validator will raise an exception if the label_object is invalid as a Cromwell label.
    """
    warnings.warn(
        "This function doesn't work for Cromwell v32 and later versions and has been deprecated, "
        "be aware of using this validator when using Cromwell v32(+). Check "
        "https://cromwell.readthedocs.io/en/stable/Labels/ for details.",
        PendingDeprecationWarning,
    )

    err_msg = ''

    if isinstance(label_object, str):
        label_object = json.loads(label_object)
    elif isinstance(label_object, bytes):
        label_object = json.loads(label_object.decode('utf-8'))
    elif isinstance(label_object, io.BytesIO):
        label_object = json.loads(label_object.getvalue())

    for label_key, label_value in label_object.items():
        err_msg += _content_checker(_CROMWELL_LABEL_KEY_REGEX, label_key)
        err_msg += _content_checker(_CROMWELL_LABEL_VALUE_REGEX, label_value)
        err_msg += _length_checker(_CROMWELL_LABEL_LENGTH, label_key)
        err_msg += _length_checker(_CROMWELL_LABEL_LENGTH, label_value)

    if err_msg != '':
        raise ValueError(err_msg)


def prepare_workflow_manifest(
    wdl_file: Union[str, io.BytesIO],
    inputs_files: Union[List[Union[str, io.BytesIO]], str, io.BytesIO] = None,
    options_file: Union[str, io.BytesIO] = None,
    dependencies: Union[str, List[str], io.BytesIO] = None,
    label_file: Union[str, io.BytesIO] = None,
    collection_name: str = None,
    on_hold: bool = False,
) -> Dict[str, Union[io.BytesIO, str]]:
    """Prepare the submission manifest for a workflow submission.

    Args:
        wdl_file: The workflow source file to submit for execution. Could be either the path
            to the file (str) or the file content in io.BytesIO.
        inputs_files: The input data in JSON
            format. Could be either the path to the file (str) or the file content in io.BytesIO. This could also
            be a list of unlimited input file paths/contents, each of them should have a type of
            Union[str, io.BytesIO].
        options_file: The Cromwell options file for workflows. Could be either
            the path to the file (str) or the file content in io.BytesIO.
        dependencies: Workflow dependency files. Could be the path to
            the zipped file (str) containing dependencies, a list of paths(List[str]) to all dependency files to be
            zipped or a zipped file in io.BytesIO.
        label_file: A collection of key/value pairs for workflow labels in JSON
                format, could be either the path to the JSON file (str) or the file content in io.BytesIO.
        collection_name: Collection in SAM that the workflow should belong to, if use CaaS.
        on_hold: Whether to submit the workflow in "On Hold" status.

    Returns:
        workflow_manifest: A dictionary representing the workflow manifest ready for workflow submission.

    Raises:
        ValueError: If a str ing of path to the dependencies is given but not endswith ".zip".
    """
    workflow_manifest = {}

    # Compose WDL source file
    workflow_manifest['workflowSource'] = _download_to_BytesIO_if_string(wdl_file)

    # Compose WDL inputs
    if inputs_files:
        if not isinstance(inputs_files, list):
            inputs_files = [inputs_files]

        for idx, inputs_file in enumerate(inputs_files):
            if idx == 0:
                # Compose WDL inputs 1
                input_file_key = 'workflowInputs'
            else:
                # Compose other WDL inputs (from 2 - many)
                input_file_key = 'workflowInputs_{X}'.format(X=idx + 1)

            workflow_manifest[input_file_key] = _download_to_BytesIO_if_string(
                inputs_file
            )

    # Compose WDL options
    if options_file:
        workflow_manifest['workflowOptions'] = _download_to_BytesIO_if_string(
            options_file
        )

    # Compose WDL labels
    if label_file:
        workflow_manifest['labels'] = _download_to_BytesIO_if_string(label_file)

    # Compose WDL dependencies
    if dependencies:
        if isinstance(dependencies, list):
            if len(dependencies) == 1 and dependencies[0].endswith('.zip'):
                # when a single zip file is provided in a list
                zip_file = _download_to_BytesIO_if_string(dependencies[0])
            else:
                # when a single wdl file or multiple wdl files are provided in a list
                zip_file = make_zip_in_memory(download_to_map(dependencies))
        elif isinstance(dependencies, str) and not dependencies.endswith('.zip'):
            # when a single file is provided as a string but not zipped
            raise ValueError(
                'The dependency file path must point to a ".zip" file! Or you may want to provide a list of WDL file(s).'
            )
        else:
            # when a single zip file is provided as a string
            zip_file = _download_to_BytesIO_if_string(dependencies)
        workflow_manifest['workflowDependencies'] = zip_file

    # Compose collection name (if use CaaS)
    if collection_name:
        workflow_manifest['collectionName'] = collection_name

    # Compose the On Hold switch for workflow submission
    workflow_manifest['workflowOnHold'] = json.dumps(on_hold)

    return workflow_manifest


def _download_to_BytesIO_if_string(file: Union[str, io.BytesIO]) -> io.BytesIO:
    """Download a file if given a string of the file path or return the input if it's in io.BytesIO.

    Args:
        file: A string of the path to the file or the file content in io.BytesIO.

    Returns:
        File content in io.BytesIO.

    Raises:
        TypeError: If the input is not a str nor io.BytesIO.
    """
    # TODO: add validation for JSON files
    if isinstance(file, str):
        return io.BytesIO(download(file))
    elif isinstance(file, io.BytesIO) or not file:
        return file
    else:
        raise TypeError('Please make sure to pass in Union[str, io.BytesIO] types!')


def compose_oauth_options_for_jes_backend_cromwell(
    auth: CromwellAuth,
    cromwell_options_file: io.BytesIO = None,
    execution_bucket: str = None,
) -> io.BytesIO:
    """Append special options that are required by JES(Google Job Execution Service) backend Cromwell.

    This helper function will append special options that are required by JES(Google Job Execution Service)
    backend Cromwell/Cromwell-as-a-Service to the default workflow options. Note: These options only work
    with Cromwell instances that use the Google Cloud Backend and allow user-service-account authentication.

    Args:
        auth: authentication class holding auth information to a Cromwell server.
        cromwell_options_file: Optional, contents of the options for a workflow in BytesIO format.
            if not specified, this function will create an empty option stream and add the necessary keys to it.
        execution_bucket: Optional, the Google CLoud Bucket that Cromwell will use to output
            execution results and store temporary scripts. If not specified, it will use
            'gs://{google_project}-cromwell-execution/caas-cromwell-executions' by default.

    Returns:
        options_stream: BytesIO object of the updated workflow options with the required auth fields.
    """
    if not cromwell_options_file:
        cromwell_options_file = io.BytesIO(json.dumps({}).encode())

    # using `getvalue()` here so we don't have to seek back to the beginning if we need the value again
    options_json = json.loads(cromwell_options_file.getvalue())
    google_project = auth.service_key_content['project_id']

    options_json.update(
        {
            'jes_gcs_root': execution_bucket
            or f'gs://{google_project}-cromwell-execution/caas-cromwell-executions',
            'google_project': google_project,
            'user_service_account_json': json.dumps(auth.service_key_content),
            'google_compute_service_account': auth.service_key_content['client_email'],
        }
    )
    options_stream = io.BytesIO(json.dumps(options_json).encode())
    return options_stream
